﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using STFXPdfDotNET;
using System.IO;
using System.Windows.Forms;
using Ionic.Zip;
using STFXmlParser;
using System.Drawing;
using System.Text.RegularExpressions;
using System.ComponentModel;
using CalcHelper;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace PdfDocument
{
    public class DocFile
    {
        private string _extPath = null;

        public string zipFile = null;
        public string pdfFile = null;
        public string xmlFile = null;

        // The JS code file in the Vscode editor.
        public string vscFile = null;

        // These are the Snippet Extension's JSON files.
        private string _codeSnipFile = null;
        private string _fcnSnipFile = null;

        public byte[] pdfBuffer = null;
        public PdfHandle hPdf = null;

        /*
         * ******************************************************
         * 
         * This area involves the displaying of what boxes are
         * controlled in the JS code file, like calcs for example.
         * Also are lists for the function declarations and code
         * comment markers encountered in the pass-thru the JS file.
         * 
         * Types of 'controlled' boxes. These values are
         * indices into the Pdf classe's ctrlColors Brush
         * array, used in ShowPage().
         */
        public int chkBoxCtrl = 0;
        public int blueField = 1;
        public int dataflow = 2;
        public int calcs = 3;

        // Every box on the document except for box1_65535.
        public List<string> docBoxList = new List<string>();

        // We default to the new box format. Ex: box393
        public bool newBoxFormat = true;

        /*
         * Controlled Box State Lists: These are lists representing
         * the current controlled state boxes per type of controlled box.
         * 
         * We get the difference between these lists and the lists set
         * by the latest run of ProcessJSFile() 
         * [see 'Controlled Box Current Lists' below] to determine
         * which boxes are no longer being controlled and which ones
         * are now being controlled by the document's code.
         */
        // Check box control:
        public List<string> cbcStateBoxes = new List<string>();

        // Blue field contol:
        List<string> bfcStateBoxes = new List<string>();

        // Dataflow:
        List<string> dfStateBoxes = new List<string>();

        List<string> calcStateBoxes = new List<string>();

        /*
         * *****************
         * Controlled Box Current Lists: These are lists representing
         * the current controlled boxes per type of controlled box.
         * 
         * These are set by ProcessJSFile() in this class.
         */
        // Check box control:
        List<string> cbcBoxes = new List<string>();

        // Blue field contol:
        List<string> bfcBoxes = new List<string>();

        // Dataflow:
        List<string> dfBoxes = new List<string>();

        List<string> calcBoxes = new List<string>();

        /*
         * Stores the Checkbox Control box sets each
         * time Calc Helper does a refresh.
         */
        public string vscodeChkBoxSets = "";

        /*
        * These are the state lists for the other dynamic
        * document's code components: Code markers which
        * are basically comments that demarcate regions of
        * code by functionality; function declarations;
        * and orphan boxes found during the vscFile pass
        * thru performed by ProcessJSFile() in this class.
        */
        public List<string> orphanStateBoxes = new List<string>();
        public List<string> docStateFunctions = new List<string>();
        public List<string> comStateMarkers = new List<string>();

        /*
         * This list is a static list which we set once. Every
         * code file has these arrays empty or not. Storing this
         * way so I do not need to use a regex in going to these
         * locations in the code file, storing what I actually
         * find for example "var calc_controlled" as opposed to
         * "var   calc_controlled".
         * 
         * There is a corresponding list on the Code Navigator
         * tab page, ctrlArrayList, that shows the 'human readable'
         * version of the array elements, we initialize that list
         * from this list during the ToolBelt constructor.
         */
        public List<string> ctrlArrays = new List<string>();

        /*
         * These are the current lists based on the current 
         * refresh pass thru.
         */
        public List<string> orphanBoxes = new List<string>();
        public List<string> docFunctions = new List<string>();
        public List<string> comMarkers = new List<string>();

        /*
         * These are for the Code Navigator tool. We use
         * BindingLists so we can easily dynamically refresh
         * the Code Navigator's lists when we Refresh Calc Helper
         * with the latest dynamic info from the code file.
         */
        public BindingList<string> codeNavOrphanBoxes = new BindingList<string>();
        public BindingList<string> codeNavDocFcns = new BindingList<string>();
        public BindingList<string> codeNavComMarkers = new BindingList<string>();

        /*
         * These dictionarys are for the Snippet Picker tool.
         * They live here because the Toolbelt can be opened and closed
         * during the liftime of Calc Helper.
         * 
         * This dictionary is for the Snippet Picker tool.
         */
        public Dictionary<string, SnippetPicker> snipPickDict = new Dictionary<string, SnippetPicker>();

        // This dictionary is for the Function Picker tool.
        public SortedDictionary<string, FunctionPicker> fcnSnipPickDict = new SortedDictionary<string, FunctionPicker>();

        public ToolBelt tb;

        /*
         * ****************************************************** 
         */

        public DocFile(string zip_File, string vsc_File, string codeSnip_File, string fcnSnip_File)
        {
            zipFile = zip_File;
            vscFile = vsc_File;
            _codeSnipFile = codeSnip_File;
            _fcnSnipFile = fcnSnip_File;
        }

        public string ExtPath
        {
            get { return _extPath; }
            set { _extPath = value; }
        }

        /*
         * This determines the box formatting being used by the
         * CalcHelper document's code. 'true' indicates the 
         * new box format is being used (box12) and 'false'
         * indicates the old box format (box12_14). We do not
         * use cid boxes to determine box format. If there 
         * are no boxes in docBoxList or if there are only
         * cid boxes in docBoxList [probably NEVER happen]
         * then we default to the new box format which might
         * not make sense, but no harm since CalcHelper can
         * not do anything with the cid boxes except maybe
         * select them or send them on to Visual Studio Code.
         * 
         * Called From:
         *      SetupPdf1() - Methods.cs.
         */
        public void SetBoxFormat()
        {
            // We default to the new box format - see up above.
            Regex oldFormat = new Regex(@"^box\d+_\d+$");

            if (docBoxList.Count > 0)
            {
                foreach (string box in docBoxList)
                {
                    if (!box.StartsWith("cid"))
                    {
                        // Only need to check the first non cid box.
                        if (oldFormat.IsMatch(docBoxList[0]))
                        {
                            newBoxFormat = false;
                        }
                        break;
                    }
                }
            }
        }

        /*
         * This retrieves the Snippet Extension's code snippet info
         * and inserts it into the snipPickDict for the 'Snippet
         * Picker' tool.
         * 
         * We sort the lists for each key category based on the prefix
         * as this is what Intellisense does in Vscode, so people will
         * have the same 'visual experience' in viewing the Snippet
         * Picker's prefix list.
         * 
         * Called From:
         *      SetupPdf1()
         */
        public void GetJSONCodeSnippets()
        {
            int hyphenIndex;
            string catKey = "";
            string catMapLine = "";
            string errMsg = "";
            string json = "{";
            List<string> tmpPrefixList;
            List<string> tmpNameList;
            List<string> tmpDescList;
            List<string> tmpBodyList;
            bool readCatMap = true;
            bool catMapStart = false;
            JObject obj = null;
            Regex catMap = new Regex(@"\s+""\/([A-Za-z0-9]+)-\/"":\s+""(.+)"",");
            Match match;
            Snippet snippet;
            SnippetPicker snippetPicker;

            try
            {
                if (!string.IsNullOrEmpty(_codeSnipFile) && File.Exists(_codeSnipFile))
                {
                    using (StreamReader sr = new StreamReader(_codeSnipFile))
                    {
                        /*
                         * Read the category map region first, and setup
                         * the dictionary.
                         */
                        while (readCatMap == true)
                        {
                            catMapLine = sr.ReadLine();

                            if (catMapLine == null)
                            {
                                // Hmmmm...
                                errMsg = string.Format("DocFile-GetJSONCodeSnippets() Error:\n\nCould not read the 'Category Mapper' section in the Snippet Extension file:\n\n{0}\n\nYou will not be able to insert code snippets from Calc Helper.", _codeSnipFile);
                                break;
                            }
                            else if (catMapLine.TrimStart().StartsWith("\"/Cat. Map-Start/\":"))
                            {
                                catMapStart = true;
                            }
                            else if (catMapLine.TrimStart().StartsWith("\"/Cat. Map-Stop/\":"))
                            {
                                readCatMap = false;
                            }
                            else if (catMapStart == true)
                            {
                                match = catMap.Match(catMapLine);

                                if (match.Success)
                                {
                                    if (!snipPickDict.ContainsKey(match.Groups[1].Value))
                                    {
                                        snippetPicker = new SnippetPicker(match.Groups[2].Value);
                                        snipPickDict.Add(match.Groups[1].Value, snippetPicker);
                                    }
                                }
                            }
                        }

                        if (errMsg.Length == 0)
                        {
                            json += sr.ReadToEnd();

                            // Preserve newlines:
                            json = json.Replace("\\n", "\\\\n");

                            obj = JObject.Parse(json);
                        }
                    }

                    if (errMsg.Length == 0)
                    {
                        if ((snipPickDict.Count == 0) || (obj == null) || (obj.Count == 0))
                        {
                            errMsg = string.Format("DocFile-GetJSONCodeSnippets() Error:\n\nThere was an error processing the Snippet Extension JSON file:\n\n{0}\n\nYou will not be able to insert code snippets from Calc Helper.", _codeSnipFile);
                        }
                        else
                        {
                            /*
                             * Now loop thru the snippet JTokens and meter out each snippet
                             * to the proper dictionary map category.
                             */
                            foreach (KeyValuePair<string, JToken> subObj in (JObject)obj)
                            {
                                snippet = subObj.Value.ToObject<Snippet>();
                                hyphenIndex = subObj.Key.IndexOf('-');

                                if (hyphenIndex < 0)
                                {
                                    errMsg = string.Format("DocFile-GetJSONCodeSnippets() Error:\n\nThere was an error processing the Snippet Extension JSON file:\n\n{0}\n\nIll formed snippet name {1}\n\nYou will not be able to insert code snippets from Calc Helper.", _codeSnipFile, subObj.Key);

                                    break;
                                }
                                else
                                {
                                    catKey = subObj.Key.Substring(0, hyphenIndex);

                                    if (snipPickDict.TryGetValue(catKey, out snippetPicker) == true)
                                    {
                                        snippetPicker.snipNames.Add(subObj.Key);
                                        snippetPicker.snipPrefixes.Add(snippet.prefix);
                                        snippetPicker.snipDescs.Add(snippet.description);
                                        snippetPicker.snipBodys.Add(string.Join(Environment.NewLine, snippet.body));
                                    }
                                    else
                                    {
                                        errMsg = string.Format("DocFile-GetJSONCodeSnippets() Error:\n\nThere was an error processing the Snippet Extension JSON file:\n\n{0}\n\nCode snippet category {1} was not in the code snippet dictionary.\n\nYou will not be able to insert code snippets from Calc Helper.", _codeSnipFile, catKey);

                                        break;
                                    }
                                }
                            }

                            /*
                             * Lastly, sort the lists in each dictionary category based
                             * on the snippet prefixes.
                             */
                            if (errMsg.Length == 0)
                            {
                                foreach (KeyValuePair<string, SnippetPicker> sp in snipPickDict)
                                {
                                    // Verify that all lists sizes are equal.
                                    if ((sp.Value.snipNames.Count == sp.Value.snipPrefixes.Count) && (sp.Value.snipNames.Count == sp.Value.snipDescs.Count) && (sp.Value.snipNames.Count == sp.Value.snipBodys.Count))
                                    {
                                        // Copy lists to tmpLists and sort on the dictionary's prefix list.
                                        tmpPrefixList = new List<string>(sp.Value.snipPrefixes);
                                        tmpNameList = new List<string>(sp.Value.snipNames);
                                        tmpDescList = new List<string>(sp.Value.snipDescs);
                                        tmpBodyList = new List<string>(sp.Value.snipBodys);

                                        sp.Value.snipNames.Clear();
                                        sp.Value.snipDescs.Clear();
                                        sp.Value.snipBodys.Clear();

                                        sp.Value.snipPrefixes.Sort();

                                        int index;

                                        foreach (string prefix in sp.Value.snipPrefixes)
                                        {
                                            index = tmpPrefixList.IndexOf(prefix);

                                            sp.Value.snipNames.Add(tmpNameList[index]);
                                            sp.Value.snipDescs.Add(tmpDescList[index]);
                                            sp.Value.snipBodys.Add(tmpBodyList[index]);
                                        }
                                    }
                                    else
                                    {
                                        errMsg = string.Format("DocFile-GetJSONCodeSnippets() Error:\n\nThere was an error processing the Snippet Extension JSON file:\n\n{0}\n\nThe lists sizes in the code snippet category {1} were not equal.\n\nYou will not be able to insert code snippets from Calc Helper.", _codeSnipFile, sp.Key);

                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
                else
                {
                    errMsg = string.Format("DocFile-GetJSONCodeSnippets() Error:\n\nCalc Helper can not locate this Snippet Extension JSON file:\n\n{0}\n\nYou will not be able to insert code snippets from Calc Helper.", _codeSnipFile);
                }

                if (errMsg.Length > 0)
                {
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                    {
                        mainMsg.ShowDialog(Form1.ActiveForm);
                    }

                    if ((snipPickDict != null) && (snipPickDict.Count > 0))
                    {
                        snipPickDict.Clear();
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "DocFile-GetJSONCodeSnippets():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * This retrieves the Snippet Extension's function snippet info
         * and inserts it into the fcnSnipPickDict for the 'Function
         * Picker' tool.
         * 
         * This is a sorted dictionary [by key] so I got off easy here.
         * 
         * Called From:
         *      SetupPdf1()
         */
        public void GetJSONFcnSnippets()
        {
            int hyphenIndex;
            string fcnType = "";
            string fcnKey = "";
            string noteLine = "";
            string errMsg = "";
            string json = "{";
            bool readNoteSection = true;
            bool addToDict = false;
            JObject obj = null;
            Snippet snippet;
            FunctionPicker functionPicker;

            try
            {
                if (!string.IsNullOrEmpty(_fcnSnipFile) && File.Exists(_fcnSnipFile))
                {
                    using (StreamReader sr = new StreamReader(_fcnSnipFile))
                    {
                        /*
                         * Read the category map region first, and setup
                         * the dictionary.
                         */
                        while (readNoteSection == true)
                        {
                            noteLine = sr.ReadLine();

                            if (noteLine == null)
                            {
                                // Hmmmm...
                                errMsg = string.Format("DocFile-GetJSONFcnSnippets() Error:\n\nCould not read the 'Note' section in the Snippet Extension file:\n\n{0}\n\nYou will not be able to insert function snippets from Calc Helper.", _fcnSnipFile);
                                break;
                            }
                            else if (noteLine.TrimStart().StartsWith("\"/Note2/\":"))
                            {
                                readNoteSection = false;
                            }
                        }

                        if (errMsg.Length == 0)
                        {
                            json += sr.ReadToEnd();

                            // Preserve newlines:
                            json = json.Replace("\\n", "\\\\n");

                            obj = JObject.Parse(json);
                        }
                    }

                    if (errMsg.Length == 0)
                    {
                        if (obj.Count == 0)
                        {
                            errMsg = string.Format("DocFile-GetJSONFcnSnippets() Error:\n\nThere was an error processing the Snippet Extension JSON file:\n\n{0}\n\nYou will not be able to insert function snippets from Calc Helper.", _fcnSnipFile);
                        }
                        else
                        {
                            /*
                             * Now loop thru the snippet JTokens and meter out each snippet
                             * to the proper dictionary map category, the key in this dictionary
                             * is the function name part of the snippet name. So for example,
                             * a snippet named "Fcn-CheckDigit" would have this as a key:
                             * "CheckDigit".
                             */
                            foreach (KeyValuePair<string, JToken> subObj in (JObject)obj)
                            {
                                snippet = subObj.Value.ToObject<Snippet>();
                                hyphenIndex = subObj.Key.IndexOf('-');

                                if (hyphenIndex < 0)
                                {
                                    errMsg = string.Format("DocFile-GetJSONFcnSnippets() Error:\n\nThere was an error processing the Snippet Extension JSON file:\n\n{0}\n\nIll formed snippet name {1}\n\nYou will not be able to insert function snippets from Calc Helper.", _fcnSnipFile, subObj.Key);

                                    break;
                                }
                                else
                                {
                                    fcnType = subObj.Key.Substring(0, hyphenIndex);

                                    if ((hyphenIndex + 1) < subObj.Key.Length)
                                    {
                                        fcnKey = subObj.Key.Substring(hyphenIndex + 1);

                                        // We need to know what type we have:
                                        if (fcnType == "Fcn")
                                        {
                                            if (fcnSnipPickDict.TryGetValue(fcnKey, out functionPicker) == false)
                                            {
                                                addToDict = true;
                                                functionPicker = new FunctionPicker();
                                            }

                                            functionPicker.fcnName = subObj.Key;
                                            functionPicker.fcnPrefix = snippet.prefix;
                                            functionPicker.fcnDesc = snippet.description;
                                            functionPicker.fcnBody = string.Join(Environment.NewLine, snippet.body);
                                        }
                                        else if (fcnType == "FcnCall")
                                        {
                                            if (fcnSnipPickDict.TryGetValue(fcnKey, out functionPicker) == false)
                                            {
                                                addToDict = true;
                                                functionPicker = new FunctionPicker();
                                            }

                                            functionPicker.fcnCallName = subObj.Key;
                                            functionPicker.fcnCallPrefix = snippet.prefix;
                                            functionPicker.fcnCallDesc = snippet.description;
                                            functionPicker.fcnCallBody = string.Join(Environment.NewLine, snippet.body);
                                        }
                                        else
                                        {
                                            errMsg = string.Format("DocFile-GetJSONFcnSnippets() Error:\n\nThere was an error processing the Snippet Extension JSON file:\n\n{0}\n\nIll formed snippet name {1}\n\nSnippet name must start with either 'Fcn-' or 'FcnCall-'.\n\nYou will not be able to insert function snippets from Calc Helper.", _fcnSnipFile, subObj.Key);

                                            break;
                                        }

                                        if (addToDict == true)
                                        {
                                            fcnSnipPickDict.Add(fcnKey, functionPicker);
                                            addToDict = false;
                                        }
                                    }
                                    else
                                    {
                                        errMsg = string.Format("DocFile-GetJSONFcnSnippets() Error:\n\nThere was an error processing the Snippet Extension JSON file:\n\n{0}\n\nIll formed snippet name {1}\n\nYou will not be able to insert function snippets from Calc Helper.", _fcnSnipFile, subObj.Key);

                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
                else
                {
                    errMsg = string.Format("DocFile-GetJSONFcnSnippets() Error:\n\nCalc Helper can not locate this Snippet Extension JSON file:\n\n{0}\n\nYou will not be able to insert function snippets from Calc Helper.", _fcnSnipFile);
                }

                if (errMsg.Length > 0)
                {
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, errMsg))
                    {
                        mainMsg.ShowDialog(Form1.ActiveForm);
                    }

                    if ((fcnSnipPickDict != null) && (fcnSnipPickDict.Count > 0))
                    {
                        fcnSnipPickDict.Clear();
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "DocFile-GetJSONFcnSnippets():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }


        /*
         * Called From:
         * 
         * refreshBtn_Click().
         * 
         * Set the Code Navigator dynamic lists [Tools tab page]:
         * 
         * We say dynamic because these lists will show the user
         * what the Refresh operation just found in the document's
         * code file, like what function declarations were encountered.
         * The user may then add a new function then save the code file
         * and again click the Refresh button since it will now be green
         * and will see that the function list will now have one more
         * item in it. Now we want to do this even if the Tool form is
         * not open, these lists are updated regardless.
         * 
         * However if the Code Navigator tab page is visible we also want
         * to make sure the selected items in each list behave properly,
         * we try to restore each list's selection to where it was before
         * we updated the list IF there was a selected item.
         * 
         * For the lists that we will change we need to turn off
         * the selection mode to avoid the list automatically
         * selecting the first item after adjusting it.
         */
        public void SetCodeNavLists()
        {
            List<string> diffList1 = new List<string>();
            List<string> diffList2 = new List<string>();

            // Do we need to change this list?
            diffList1 = docFunctions.Except(docStateFunctions).ToList();
            diffList2 = docStateFunctions.Except(docFunctions).ToList();

            if ((diffList1.Count > 0) || (diffList2.Count > 0))
            {
                if (tb != null)
                {
                    tb.CodeNavListSelectMode(CalcHelper.ToolBelt.codeNavList.docFcns, false);
                }

                codeNavDocFcns.Clear();

                foreach (string fcn in docFunctions)
                {
                    codeNavDocFcns.Add(Regex.Replace(fcn, @"(^function\s+)|(\s*{$)", ""));
                }

                docStateFunctions = docFunctions.ToList();

                if (tb != null)
                {
                    tb.CodeNavListSelectMode(CalcHelper.ToolBelt.codeNavList.docFcns, true);
                    tb.SetCodeNavListSel(CalcHelper.ToolBelt.codeNavList.docFcns);
                }
            }

            // Do we need to change this list?
            diffList1 = comMarkers.Except(comStateMarkers).ToList();
            diffList2 = comStateMarkers.Except(comMarkers).ToList();

            if ((diffList1.Count > 0) || (diffList2.Count > 0))
            {
                if (tb != null)
                {
                    tb.CodeNavListSelectMode(CalcHelper.ToolBelt.codeNavList.comMrkrs, false);
                }

                codeNavComMarkers.Clear();

                foreach (string com in comMarkers)
                {
                    codeNavComMarkers.Add(com);
                }

                comStateMarkers = comMarkers.ToList();

                if (tb != null)
                {
                    tb.CodeNavListSelectMode(CalcHelper.ToolBelt.codeNavList.comMrkrs, true);
                    tb.SetCodeNavListSel(CalcHelper.ToolBelt.codeNavList.comMrkrs);
                }
            }

            // Do we need to change this list?
            diffList1 = orphanBoxes.Except(orphanStateBoxes).ToList();
            diffList2 = orphanStateBoxes.Except(orphanBoxes).ToList();

            if ((diffList1.Count > 0) || (diffList2.Count > 0))
            {
                if (tb != null)
                {
                    tb.CodeNavListSelectMode(CalcHelper.ToolBelt.codeNavList.orphBoxes, false);
                }

                codeNavOrphanBoxes.Clear();

                foreach (string orphan in orphanBoxes)
                {
                    codeNavOrphanBoxes.Add(orphan);
                }

                orphanStateBoxes = orphanBoxes.ToList();

                if (tb != null)
                {
                    tb.CodeNavListSelectMode(CalcHelper.ToolBelt.codeNavList.orphBoxes, true);
                    tb.SetCodeNavListSel(CalcHelper.ToolBelt.codeNavList.orphBoxes);
                }
            }
        }

        /*
         * Called From:
         * 
         *  SetupPdf1() - While opening a file for display.
         *  refreshBtn_Click() - While refreshing Calc Helper with
         *      info from a document's Javascript file.
         *
         * We read the document's .JS code file and gather lists of boxes
         * that are Check Box Controlled, Dataflowed, Calculated, Blue
         * Field controlled, Orphan boxes. We also build lists for
         * Comment code markers - they delineate regions of code for certain
         * purposes, like posting for example. We also build a list of any
         * function declarations encountered.
         * 
         * We also grab a copy of the raw Checkbox Control set for use when
         * we are sending checkbox selections to Checkbox Control.
         * 
         * refresh: This will be 'false' when opening a document for displaying
         *      in Calc Helper. It will be 'true' when refreshing as a result
         *      of the document's code file being saved in the editor.
         *      
         * pa1: This is the PageArray for pdf1. We need this to set the box fields
         *      on pages that contain boxes that are now being controlled or no
         *      longer being controlled.
         *      
         * pdf1: This is for displaying a pdf page, the main display panel.
         * 
         * pa2: This is the PageArray for pdf2. We will make sure this is not null,
         *      this would be if the screen was not split. If it is split then we
         *      need to make the same control updates to the same fields.
         *      
         * pdf2: This is for displaying a pdf page, the second display panel provided
         *         we have a split screen, else this will be null.
         */
        public bool ProcessJSFile(bool refresh, ref PageArray pa1, ref Pdf pdf1, ref PageArray pa2, ref Pdf pdf2)
        {
            bool result = true;
            bool getChkBoxes = false;
            bool chkBoxesDone = false;
            bool blueFieldDone = false;
            bool dfDone = false;
            bool globalVarsDone = false;

            string[] jsCode;
            string msgStr = "";
            string line = "";
            string codeStr = "";

            int dfFlagCnt = 0;

            MatchCollection matches;
            Match match;

            Regex dfProfileRef = new Regex(@"(entity|equifax|fiduciary|individual|preparer)\.");
            Regex dfSetValue1Box;
            Regex dfSetValue2Box;

            // These are comment markers - they delineate code sections.
            Regex dfEnd = new Regex(@"^//(((no|stop|end(of)?)dataflow)|(dataflow(end|stop)))$");
            Regex notes = new Regex(@"^//relevantworkflow/spec/formattingnotes$");
            Regex posting = new Regex(@"^//posting(stuff)?$");
            Regex methods = new Regex(@"^//callmethods/tables$");
            Regex formCalcs = new Regex(@"^//formcalcs$");
            Regex barcode1D = new Regex(@"^//1dbarcode((\(s\))|s)?$");
            Regex barcode2D = new Regex(@"^//2dbarcode((\(s\))|s)?$");
            Regex scanLine = new Regex(@"^//scanline((\(s\))|s)?$");
            Regex scanBand = new Regex(@"^//scanband((\(s\))|s)?$");
            Regex globalVars = new Regex(@"^//globalvariables$");

            // Checks for function declarations:
            Regex function = new Regex(@"^(function\s+\w+)\s*\((\s*\w+\s*(,\s*\w+\s*)*)?\)\s*\{$");

            // Check for a box assignment.
            Regex boxAssign;

            // Used in determing boxes added or deleted from control.
            List<string> diffList = new List<string>();

            List<string> maybeDfBoxes = new List<string>();

            /*
             * All unique boxes found in the document's code.
             * Used in building the orphanBoxes list.
             */
            List<string> codeBoxes = new List<string>();

            if (newBoxFormat == true)
            {
                dfSetValue1Box = new Regex(@"\s*SFForms\.StorageManager\.setValue\(\s*formInstanceId\s*,\s*[""'](box\d+)[""'](?!\s*,\s*box\d+)");
                dfSetValue2Box = new Regex(@"\s*SFForms\.StorageManager\.setValue\(\s*formInstanceId\s*,.+\?\s*[""'](box\d+)[""']\s*:\s*[""'](box\d+)[""'](?!\s*\)?,\s*box\d+)");
                boxAssign = new Regex(@"^(box\d+)\s*=(?!=).+");
            }
            else
            {
                dfSetValue1Box = new Regex(@"\s*SFForms\.StorageManager\.setValue\(\s*formInstanceId\s*,\s*[""'](box\d+_\d+)[""'](?!\s*,\s*box\d+_\d+)");
                dfSetValue2Box = new Regex(@"\s*SFForms\.StorageManager\.setValue\(\s*formInstanceId\s*,.+\?\s*[""'](box\d+_\d+)[""']\s*:\s*[""'](box\d+_\d+)[""'](?!\s*\)?,\s*box\d+_\d+)");
                boxAssign = new Regex(@"^(box\d+_\d+)\s*=(?!=).+");
            }

            // Used when sending checbox sets to CBC.
            vscodeChkBoxSets = "";

            /*
             * Caller has ensured pa1 and pdf1 are not null. We will need
             * to check pa2 and pdf2. These are actually referenced in
             * ProcessBoxCtrl().
             */

            if (refresh)
            {
                /*
                 * Clear out the various current lists that will be
                 * refreshed with new information.
                 */
                cbcBoxes.Clear();
                bfcBoxes.Clear();
                dfBoxes.Clear();
                calcBoxes.Clear();
                orphanBoxes.Clear();
                docFunctions.Clear();
                comMarkers.Clear();
            }

            try
            {
                jsCode = File.ReadAllText(vscFile).Split(new string[] { "\r\n", "\n" }, StringSplitOptions.RemoveEmptyEntries);

                for (int i = 0; i < jsCode.Length; i++)
                {
                    /*
                     * Look for the Check Box Control box array and get any boxes
                     * in the array. This is in every file, empty or not. So we will
                     * just focus on this first.
                     */
                    if (!chkBoxesDone)
                    {
                        if (getChkBoxes)
                        {
                            getChkBoxes = false;
                            chkBoxesDone = true;

                            matches = Regex.Matches(jsCode[i], (newBoxFormat == true ? @"box\d+" : @"box\d+_\d+"));

                            /*
                             * Grab any boxes and put in codeStr, we will use this to generate
                             * the orphan box list after we read all lines. Grab the whole line
                             * for now. Also get a copy for when we send selections to CBC.
                             */
                            if (matches.Count > 0)
                            {
                                codeStr += jsCode[i];
                                vscodeChkBoxSets = jsCode[i];
                            }

                            foreach (Match chkBox in matches)
                            {
                                if (docBoxList.Contains(chkBox.Value))
                                {
                                    cbcBoxes.Add(chkBox.Value);
                                }
                            }
                        }
                        else if (jsCode[i].Replace(" ", String.Empty).Contains("functiongetCheckBoxList()"))
                        {
                            if (refresh == false)
                            {
                                /*
                                 * We know we have one ')' here. We will search for this text when
                                 * selected from the accompanying list on the Code Navigator tab page.
                                 */
                                ctrlArrays.Add(jsCode[i].Substring(0, jsCode[i].IndexOf(')') + 1).TrimStart());
                            }

                            // Next line will be the box array, empty or not.
                            getChkBoxes = true;
                        }

                        continue;
                    }

                    /*
                     * Look for the calc_controlled (aka blue field control) box array and
                     * get any boxes in the array. This is in every file, empty or not so
                     * we will next focus on just this.
                     */
                    if (!blueFieldDone)
                    {
                        if (jsCode[i].Replace(" ", String.Empty).Contains("varcalc_controlled=["))
                        {
                            if (refresh == false)
                            {
                                /*
                                 * We know we have one '=' here. We will search for this text when
                                 * selected from the accompanying list on the Code Navigator tab page.
                                 */
                                ctrlArrays.Add(jsCode[i].Substring(0, jsCode[i].IndexOf('=')).Trim());
                            }

                            // We are on the Blue Field Control box array line.
                            matches = Regex.Matches(jsCode[i], (newBoxFormat == true ? @"box\d+" : @"box\d+_\d+"));

                            // Grab any boxes and put in codeStr.
                            if (matches.Count > 0)
                            {
                                codeStr += jsCode[i];
                            }

                            foreach (Match bfBox in matches)
                            {
                                if (docBoxList.Contains(bfBox.Value))
                                {
                                    /*
                                     * Right now there is not much control over this array
                                     * so checking for human error.
                                     */
                                    if (!bfcBoxes.Contains(bfBox.Value))
                                    {
                                        bfcBoxes.Add(bfBox.Value);
                                    }
                                }
                            }

                            blueFieldDone = true;
                        }

                        continue;
                    }

                    /*
                     * The Global Vars comment 'marker' currently sits just after the 
                     * calc_controlled array - sits on its own in the marker grouping.
                     * 
                     * NOTE: We are removing #'s because they are present on the comment
                     * marker lines, several adjacent to each other, something we retained
                     * from Ruby. '*' and '-' are/have been on the 'Global Variables'
                     * comment marker line - Calcs does not dictate this. Can't remember
                     * why I have '.' in here..doh!
                     */
                    line = Regex.Replace(jsCode[i].ToLower(), @"[\s#.*-]", string.Empty);

                    if (!globalVarsDone)
                    {
                        if (globalVars.IsMatch(line))
                        {
                            comMarkers.Add(jsCode[i].Trim());
                            globalVarsDone = true;
                        }

                        continue;
                    }

                    /*
                     * Keep rolling the codeStr if not a comment line. We are also just entering
                     * the dataflow zone, noting that it may be possible that there is no
                     * dataflow.
                     */
                    if ((!line.StartsWith("//")) && ((line.Contains("box")) || (line.Contains("cid"))))
                    {
                        codeStr += jsCode[i];
                    }

                    /*
                     * Either we have dataflow in this doc or we do not, either way we are done
                     * when we: Know we have dataflow and we move out of that area; Or we find out
                     * there is no dataflow in this file.
                     */
                    if (!dfDone)
                    {
                        if (dfProfileRef.IsMatch(jsCode[i]))
                        {
                            dfFlagCnt += 1;

                            /*
                             * Trying this commented out - saw a case where the "..entity.zip..."
                             * on a myMatches line had a dataflow to a box on the same line, so
                             * we would miss it if the continue fired.
                             */
                            //continue;
                        }

                        // Regular dataflow (to 1 box):
                        match = dfSetValue1Box.Match(jsCode[i]);

                        if (match.Success)
                        {
                            if (docBoxList.Contains(match.Groups[1].Value))
                            {
                                if (dfFlagCnt > 0)
                                {
                                    /*
                                     * Several profiles can flow to the same box:
                                     */
                                    if (!dfBoxes.Contains(match.Groups[1].Value))
                                    {
                                        dfBoxes.Add(match.Groups[1].Value);
                                    }
                                }
                                /*
                                 * 07/30/19 Defaulting to dataflow if I see a match and dfFlagCnt is positive.
                                 * Not decrementing the flag, a few tests showed that I missed some dataflow
                                 * that looked like "....setValue(formInstanceId, 'box1_5', myMatches[2]);"
                                 * this was in a group of 3 lines of dataflow based on an ssn match - 
                                 * dataflowing a ssn to 3 boxes based on a single match. The prior way caught
                                 * the first box but missed the last 2 because the flag was reset to zero and
                                 * was not incremented because dfProfileRef did not match again.
                                 */
                                //else
                                //{
                                //    // Strange situation - hopefully will not happen...
                                //    if (!maybeDfBoxes.Contains(match.Groups[1].Value))
                                //    {
                                //        maybeDfBoxes.Add(match.Groups[1].Value);
                                //    }
                                //}
                            }

                            //dfFlagCnt = 0;

                            continue;
                        }

                        /*
                         * Dataflow to 1 of 2 boxes (address collapse).
                         * Same logic as to dataflow for just one box above.
                         */
                        match = dfSetValue2Box.Match(jsCode[i]);

                        if (match.Success)
                        {
                            for (int j = 1; j < match.Groups.Count; j++)
                            {
                                if (docBoxList.Contains(match.Groups[j].Value))
                                {
                                    if (dfFlagCnt > 0)
                                    {
                                        /*
                                         * Several profiles can flow to the same box:
                                         */
                                        if (!dfBoxes.Contains(match.Groups[j].Value))
                                        {
                                            dfBoxes.Add(match.Groups[j].Value);
                                        }
                                    }
                                }
                            }

                            continue;
                        }

                        // Are we at the end of the dataflow area, or do we just don't have any dataflow?
                        if (dfEnd.IsMatch(line) || notes.IsMatch(line) || posting.IsMatch(line) || methods.IsMatch(line) || formCalcs.IsMatch(line) || barcode1D.IsMatch(line) || barcode2D.IsMatch(line) || scanLine.IsMatch(line) || scanBand.IsMatch(line))
                        {
                            dfDone = true;

                            comMarkers.Add(jsCode[i].Trim());

                            continue;
                        }

                        // Ain't no functions in dataflow Willis!
                        match = function.Match(jsCode[i].Trim());

                        if (match.Success)
                        {
                            dfDone = true;

                            docFunctions.Add(match.Groups[0].Value);

                            continue;
                        }

                        // Ain't no box assignments in dataflow either Willis!
                        match = boxAssign.Match(jsCode[i].Trim());

                        if (match.Success)
                        {
                            dfDone = true;

                            if (docBoxList.Contains(match.Groups[1].Value))
                            {
                                if (!bfcBoxes.Contains(match.Groups[1].Value))
                                {
                                    if (!calcBoxes.Contains(match.Groups[1].Value))
                                    {
                                        calcBoxes.Add(match.Groups[1].Value);
                                    }
                                }
                            }

                            continue;
                        }

                        continue;
                    }

                    /*
                     * From here on out we look for any code markers, functions and calc boxes.
                     * Pretty much can be a mixture of them.
                     */
                    match = boxAssign.Match(jsCode[i].Trim());

                    if (match.Success)
                    {
                        if (docBoxList.Contains(match.Groups[1].Value))
                        {
                            if (!bfcBoxes.Contains(match.Groups[1].Value))
                            {
                                if (!calcBoxes.Contains(match.Groups[1].Value))
                                {
                                    calcBoxes.Add(match.Groups[1].Value);
                                }
                            }
                        }

                        continue;
                    }

                    match = function.Match(jsCode[i].Trim());

                    if (match.Success)
                    {
                        docFunctions.Add(match.Groups[0].Value);

                        continue;
                    }

                    if (notes.IsMatch(line) || posting.IsMatch(line) || methods.IsMatch(line) || formCalcs.IsMatch(line) || barcode1D.IsMatch(line) || barcode2D.IsMatch(line) || scanLine.IsMatch(line) || scanBand.IsMatch(line))
                    {
                        comMarkers.Add(jsCode[i].Trim());
                    }
                }

                /*
                 * Now figure out the orphan box situation by comparing a box list
                 * extracted from the codeStr rolled up above to the docBoxList
                 * i.e. a list of all boxes on the document.
                 */
                if (!string.IsNullOrEmpty(codeStr))
                {
                    if (newBoxFormat == true)
                    {
                        codeBoxes = Regex.Matches(codeStr, @"(box\d+)|(cid\d+_\d+)").Cast<Match>().Select(box => box.Value).ToList();
                    }
                    else
                    {
                        codeBoxes = Regex.Matches(codeStr, @"(box\d+_\d+)|(cid\d+_\d+)").Cast<Match>().Select(box => box.Value).ToList();
                    }

                    codeBoxes = codeBoxes.Distinct().ToList();
                    orphanBoxes = codeBoxes.Except(docBoxList).ToList();
                }

                if (refresh)
                {
                    /* 
                     * Processing a document refresh [JS code file has been
                     * saved in the code editor] requires a comparison against the
                     * current state of the dynamic box lists to determine which
                     * boxes are 'new' regarding being calculated, dataflowed, etc.
                     * [we refer to this as being 'controlled'] and which boxes are
                     * no longer in that state.
                     */
                    // Check box control boxes now being controlled.
                    diffList = cbcBoxes.Except(cbcStateBoxes).ToList();

                    if (diffList.Count > 0)
                    {
                        result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, chkBoxCtrl, true);
                    }

                    if (result)
                    {
                        // Check box control boxes no longer being controlled.
                        diffList = cbcStateBoxes.Except(cbcBoxes).ToList();

                        if (diffList.Count > 0)
                        {
                            result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, chkBoxCtrl, false);
                        }
                    }

                    // Blue field control - same logic:
                    if (result)
                    {
                        diffList = bfcBoxes.Except(bfcStateBoxes).ToList();

                        if (diffList.Count > 0)
                        {
                            result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, blueField, true);
                        }
                    }

                    if (result)
                    {
                        diffList = bfcStateBoxes.Except(bfcBoxes).ToList();

                        if (diffList.Count > 0)
                        {
                            result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, blueField, false);
                        }
                    }

                    // Data flow - same logic:
                    if (result)
                    {
                        diffList = dfBoxes.Except(dfStateBoxes).ToList();

                        if (diffList.Count > 0)
                        {
                            result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, dataflow, true);
                        }
                    }

                    if (result)
                    {
                        diffList = dfStateBoxes.Except(dfBoxes).ToList();

                        if (diffList.Count > 0)
                        {
                            result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, dataflow, false);
                        }
                    }

                    // Calcs - same logic:
                    if (result)
                    {
                        diffList = calcBoxes.Except(calcStateBoxes).ToList();

                        if (diffList.Count > 0)
                        {
                            result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, calcs, true);
                        }
                    }

                    if (result)
                    {
                        diffList = calcStateBoxes.Except(calcBoxes).ToList();

                        if (diffList.Count > 0)
                        {
                            result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, calcs, false);
                        }
                    }
                }
                else
                {
                    /*
                     * Here every box is 'new' regarding calculation, check box
                     * controlled, etc.. Set the PageArray fields for controlled
                     * boxes.
                     */
                    if (cbcBoxes.Count > 0)
                    {
                        result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref cbcBoxes, chkBoxCtrl, true);
                    }

                    if (result && (bfcBoxes.Count > 0))
                    {
                        result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref bfcBoxes, blueField, true);
                    }

                    if (result && (dfBoxes.Count > 0))
                    {
                        result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref dfBoxes, dataflow, true);
                    }

                    if (result && (calcBoxes.Count > 0))
                    {
                        result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref calcBoxes, calcs, true);
                    }
                }

                if (result)
                {
                    // Store the box control state for the next refresh:
                    cbcStateBoxes = cbcBoxes.ToList();
                    bfcStateBoxes = bfcBoxes.ToList();
                    dfStateBoxes = dfBoxes.ToList();
                    calcStateBoxes = calcBoxes.ToList();
                }
            }
            catch (Exception ex)
            {
                result = false;

                MessageBox.Show(ex.ToString(), "DocFile-ProcessJSFile():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            /*
             * Safety check - check these 'done' flags and popup a warning
             * message if not set. More of a debug thing for me, but 
             * just had a problem with no controlled calcs displaying
             * except for check box control and there was dataflow and
             * calcs present, this was due to a change in the 'Global
             * Variables' comment line, '-' were replaced with '*'s.
             */
            if (chkBoxesDone == false)
            {
                msgStr = "ProcessJSFile() Error:\n\nCheckbox Control box array could not be located.\n\nThis will prevent controlled boxes from being correctly displayed.";
            }
            else if (blueFieldDone == false)
            {
                msgStr = "ProcessJSFile() Error:\n\ncalc_controlled box array could not be located.\n\nThis holds the 'bluefield control' boxes.\n\nThis will prevent controlled boxes from being correctly displayed.";
            }
            else if (globalVarsDone == false)
            {
                msgStr = "ProcessJSFile() Error:\n\n'Global Variables' comment line could not be located.\n\nThis will prevent controlled boxes from being correctly displayed.";
            }
            else if (dfDone == false)
            {
                if (dfFlagCnt > 0)
                {
                    msgStr = "ProcessJSFile() Error:\n\nIt could not be determined whether there is or is not any Dataflow.\n\nThis will prevent controlled boxes from being correctly displayed.\n\nPlease check if there is an end of Dataflow comment and add one if there is not, then save the code file and Refresh Calc Helper.";
                }
                else
                {
                    msgStr = "ProcessJSFile() Error:\n\nIt could not be determined whether there is or is not any Dataflow.\n\nThis will prevent controlled boxes from being correctly displayed.";
                }
            }

            if (msgStr.Length > 0)
            {
                MessageBox.Show(Form1.ActiveForm, msgStr, "DocFile-ProcessJSFile():", MessageBoxButtons.OK, MessageBoxIcon.Error);

                /*
                 * Extreme testing has revealed that this might lockup Calc Helper like the
                 * ctEvent docLock message used to do in Form1.cs, this was used there also.
                 */
                //using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msgStr))
                //{
                //mainMsg.ShowDialog(Form1.ActiveForm);
                //}
            }

            return result;
        }

        /*
         * Decided not to use this but keeping it around.
         * 
         * Called From:
         * 
         *  ctClient_CtClientEvent()
         *
         * We perform what ProcessJSFile() does but just for checkboxes. The
         * boxes are sent to Calc Helper from Checkbox Control, they are the
         * boxes in the Checkbox Sets list. When selecting checkboxes for 
         * the Checkbox Control tool we will receive the contents of this list
         * when it changes. Calc Helper will process the check boxes and the
         * user will be able to view the progress by looking at the document
         * in Calc Helper.
         * 
         * Arguments: 
         * 
         * chkBoxStr: The contents of the Checkbox Sets list as a string.
         *      
         * pa1: This is the PageArray for pdf1. We need this to set the checkbox
         *      fields on pages that contain checkboxes that are now being
         *      controlled or no longer being controlled.
         *      
         * pdf1: This is for displaying a pdf page, the main display panel.
         * 
         * pa2: This is the PageArray for pdf2. We will make sure this is not null,
         *      this would be if the screen was not split. If it is split then we
         *      need to make the same control updates to the same fields.
         *      
         * pdf2: This is for displaying a pdf page, the second display panel provided
         *         we have a split screen, else this will be null.
         */
        //public bool ProcessChkBoxCtrlSets(string chkBoxStr, ref PageArray pa1, ref Pdf pdf1, ref PageArray pa2, ref Pdf pdf2)
        //{
        //    bool result = true;
        //    MatchCollection matches;

        //    // Used in determing boxes added or deleted from control.
        //    List<string> diffList = new List<string>();

        //    /*
        //     * Caller has ensured pa1 and pdf1 are not null. We will need
        //     * to check pa2 and pdf2. These are actually referenced in
        //     * ProcessBoxCtrl().
        //     */

        //    // First clear the current values.
        //    cbcBoxes.Clear();

        //    matches = Regex.Matches(chkBoxStr, (newBoxFormat == true ? @"box\d+" : @"box\d+_\d+"));

        //    /*
        //     * We will make sure the boxes are on the document since this info
        //     * has come from the Checkbox Control tool and there could possibly
        //     * be orphan boxes currently in the set list.
        //     */
        //    foreach (Match chkBox in matches)
        //    {
        //        if (docBoxList.Contains(chkBox.Value))
        //        {
        //            cbcBoxes.Add(chkBox.Value);
        //        }
        //    }

        //    // Check box control boxes now being controlled.
        //    diffList = cbcBoxes.Except(cbcStateBoxes).ToList();

        //    if (diffList.Count > 0)
        //    {
        //        result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, chkBoxCtrl, true);
        //    }

        //    if (result)
        //    {
        //        // Check box control boxes no longer being controlled.
        //        diffList = cbcStateBoxes.Except(cbcBoxes).ToList();

        //        if (diffList.Count > 0)
        //        {
        //            result = ProcessBoxCtrl(ref pa1, ref pdf1, ref pa2, ref pdf2, ref diffList, chkBoxCtrl, false);

        // This has been removed from DocFile.
        //            cbcUnSelBoxes = diffList.ToList();
        //        }
        //    }

        //    if (result)
        //    {
        //        // Store the box control state for the next processing:
        //        cbcStateBoxes = cbcBoxes.ToList();
        //    }

        //    return result;
        //}

        /*
         * This sets the PageArray fields [Field.cs] to
         * indicate which boxes are being controlled or
         * are no longer being controlled. Controlled refers
         * to being calculated, blue field controlled, check
         * box controlled or being dataflowed.
         * 
         * Called From:
         * 
         *      ProcessJSFile()
         *      ProcessChkBoxCtrlSets()
         */
        public bool ProcessBoxCtrl(ref PageArray pa1, ref Pdf pdf1, ref PageArray pa2, ref Pdf pdf2, ref List<string> boxList, int boxType, bool ctrlStatus)
        {
            int faIndex;
            int curPage;
            int paPage;
            bool result = true;
            bool split = false;
            List<Field> fa1 = null;
            List<Field> fa2 = null;
            Dictionary<string, int> boxPageDict = new Dictionary<string, int>();

            /*
             * We will need to check if pa2 and pdf2 are null as
             * we may not have a split screen. If the screen is split
             * we will need to set the same fields in pa2.
             */
            if ((pdf2 != null) && (pa2 != null))
            {
                split = true;
            }

            /*
             * Caller has ensured list is not empty. Sorting just on the
             * boxes [new box format] is no guarantee you are sorting by page
             * order which is the whole intent of sorting here. The old box format
             * was sorting by page order, as the page number was part of the box
             * number but not anymore. We are trying to avoid needless
             * switching between field array fetching below.
             * 
             * The ONLY box on the document that we do not keep track of is
             * box1_65535, so we can just ask pa1 for the page number here.
             */
            foreach (string box in boxList)
            {
                if (!boxPageDict.ContainsKey(box))
                {
                    boxPageDict.Add(box, pa1.GetBoxPageNum(box));
                }
            }

            try
            {
                paPage = -1;

                foreach (KeyValuePair<string, int> kvp in boxPageDict.OrderBy(key => key.Value))
                {
                    curPage = kvp.Value;

                    if (pdf1.refreshPage == false)
                    {
                        if (pdf1.curPage == curPage)
                        {
                            pdf1.refreshPage = true;
                        }
                    }

                    if ((split == true) && (pdf2.refreshPage == false))
                    {
                        if (pdf2.curPage == curPage)
                        {
                            pdf2.refreshPage = true;
                        }
                    }

                    if (curPage != paPage)
                    {
                        pa1.GetFieldArrayRef(curPage - 1, ref fa1);
                        paPage = curPage;

                        if (fa1 == null)
                        {
                            using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("ProcessBoxCtrl() Error:\n\npa1.GetFieldArrayRef() returned null field\narray while processing {0}.", kvp.Key)))
                            {
                                mainMsg.ShowDialog(Form1.ActiveForm);
                            }

                            result = false;
                            break;
                        }

                        if (split == true)
                        {
                            pa2.GetFieldArrayRef(curPage - 1, ref fa2);

                            if (fa2 == null)
                            {
                                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("ProcessBoxCtrl() Error:\n\npa2.GetFieldArrayRef() returned null field\narray while processing {0}.\n\nRefresh changes will not be displayed for\nthe second display screen.", kvp.Key)))
                                {
                                    mainMsg.ShowDialog(Form1.ActiveForm);
                                }

                                pdf2.refreshPage = false;
                                split = false;
                            }
                        }
                    }

                    faIndex = pa1.GetBoxFieldArrayIndex(kvp.Key);

                    fa1[faIndex].ctrlType = (ctrlStatus ? boxType : -1);

                    if (split == true)
                    {
                        /*
                         * By definition the pa's and fa's are identical since
                         * they are both tied to the same document.
                        */
                        fa2[faIndex].ctrlType = fa1[faIndex].ctrlType;
                    }
                }
            }
            catch (Exception ex)
            {
                result = false;

                MessageBox.Show(ex.ToString(), "DocFile-ProcessBoxCtrl():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            return result;
        }

        public Boolean GetXmlData(ref PageArray pa)
        {
            Boolean success = false;

            if (File.Exists(xmlFile))
            {
                try
                {
                    pa = new PageArray();
                    byte[] xml = File.ReadAllBytes(xmlFile);

                    if (xml != null)
                    {
                        XmlParser parser = new XmlParser(ref pa, xml);
                        success = parser.Parse();
                    }
                }
                catch (Exception ex)
                {
                    success = false;

                    MessageBox.Show(ex.ToString(), "GetXmlData():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

            return success;
        }

        public Boolean SetPdfBuffer()
        {
            Boolean success = false;
            CSTFPdf STFPdf = new CSTFPdf();

            if (File.Exists(pdfFile))
            {
                try
                {
                    pdfBuffer = File.ReadAllBytes(pdfFile);

                    if (pdfBuffer != null)
                    {
                        hPdf = STFPdf.OpenPdfMem(pdfBuffer, "", "");
                        success = true;
                    }
                }
                catch (Exception ex)
                {
                    success = false;

                    if (hPdf != null)
                    {
                        STFPdf.ClosePdf(hPdf);
                    }

                    MessageBox.Show(ex.ToString(), "SetPdfBuffer():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

            return success;
        }

        public Boolean ExtractFiles()
        {
            string tmpPath = null;
            string ZIP_PW = "5e$ACre7ru";
            Boolean retValue = false;

            tmpPath = GetTempDirectory();

            if (!string.IsNullOrEmpty(tmpPath))
            {
                if (Directory.Exists(tmpPath))
                {
                    try
                    {
                        using (ZipFile zip = ZipFile.Read(zipFile))
                        {
                            foreach (ZipEntry e in zip)
                            {
                                if (e.FileName.EndsWith(".pdf") || e.FileName.EndsWith(".xml"))
                                {
                                    if (e.UsesEncryption == true)
                                    {
                                        e.ExtractWithPassword(tmpPath, ZIP_PW);
                                    }
                                    else
                                    {
                                        e.Extract(tmpPath);
                                    }
                                }
                            }
                        }

                        pdfFile = Path.Combine(tmpPath, "doc.pdf");
                        xmlFile = Path.Combine(tmpPath, "doc.xml");

                        // It makes no sense to continue with this document if these are missing.
                        if (File.Exists(pdfFile) && File.Exists(xmlFile))
                        {
                            retValue = true;
                            //MessageBox.Show(string.Format("Extracted To: {0}", tmpPath));
                        }
                        else
                        {
                            pdfFile = null;
                            xmlFile = null;
                        }
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(ex.ToString(), "ExtractFiles():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
            }

            _extPath = tmpPath;
            return retValue;
        }

        private string GetTempDirectory()
        {
            string path = null;

            try
            {
                path = Path.GetRandomFileName();
                path = Path.Combine(Path.GetTempPath(), path);
                Directory.CreateDirectory(path);
            }
            catch (Exception ex)
            {
                path = null;
                MessageBox.Show(ex.ToString(), "GetTempDirectory():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            return path;
        }

        public void CleanupFiles()
        {
            //MessageBox.Show(string.Format("Entered CleanupFiles() -> unzipPath: {0}", _extPath));

            if (!string.IsNullOrEmpty(_extPath))
            {
                if (Directory.Exists(_extPath))
                {
                    try
                    {
                        Directory.Delete(_extPath, true);

                        if (hPdf != null)
                        {
                            CSTFPdf STFPdf = new CSTFPdf();

                            STFPdf.ClosePdf(hPdf);
                        }
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(ex.ToString(), string.Format("CleanupFiles({0}):", _extPath), MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
                _extPath = null;
            }
        }
    }

    /*
     * As the name suggests, this is for the Tool Belt's
     * 'Snippet Picker' tool...right up to the knuckle baby!
     * 
     * catName is the 'human readable' version of the category name.
     * Category name would be "Calc", the catName would be something
     * like "Calculation Templates" which would be in the category list
     * on the Snippet Picker.
     * 
     * snipNames are the actual snippet names that Vscode knows, the user
     * does not see these, nor do they know they even exist.
     * 
     * The other components are Vscode snippet components, though I
     * spell them differently.
     */
    public class SnippetPicker
    {
        public string catName;
        public List<string> snipNames = new List<string>();
        public List<string> snipPrefixes = new List<string>();
        public List<string> snipDescs = new List<string>();
        public List<string> snipBodys = new List<string>();

        public SnippetPicker(string categoryName)
        {
            catName = categoryName;
        }
    }

    /*
     * As the name suggests, this is for the Tool Belt's
     * 'Function Picker' tool. Similar naming logic as for
     * the Snippet Picker, that is, pseudo naming off from
     * Vscode snippet components.
     */
    public class FunctionPicker
    {
        public string fcnName = "NOT ASSIGNED";
        public string fcnPrefix = "NOT ASSIGNED";
        public string fcnBody = "NOT ASSIGNED";
        public string fcnDesc = "NOT ASSIGNED";
        public string fcnCallName = "NOT ASSIGNED";
        public string fcnCallPrefix = "NOT ASSIGNED";
        public string fcnCallBody = "NOT ASSIGNED";
        public string fcnCallDesc = "NOT ASSIGNED";
    }

    /*
     * This class is for both the code snippets and the
     * function snippets from the Snippet Extension's
     * 2 JSON files.
     */
    public class Snippet
    {
        public string prefix { get; set; }
        public List<string> body { get; set; }
        public string description { get; set; }
    }

}